// GBList.c

#include "GBList.hpp"
#ifdef USE_OLD_GBLIST
#include "MyOstream.hpp"
#include "RecordHere.hpp"
//#ifndef DEBUGDEEPCOPY
//#define DEBUGDEEPCOPY
//#endif


// CAN NOT BE INLINED :-(
template<class T> void GBList<T>::push_back(const T & a) {
  if (rep == NULL) {
    RECORDHERE(rep = new GBListRep<T>();)
    if(rep==NULL) DBG();
  } else if (rep->count() != 1)  {// More than one instance
    DeepCopy();
  };
  rep->push_back(a);
}

// CAN NOT BE INLINED :-(
template<class T>
void GBList<T>::addElement(const T & a, int index) {
  if (rep == NULL) {
    RECORDHERE(rep = new GBListRep<T>();)
    if(rep==NULL) DBG();
  } else if (rep->count() != 1)  {// More than one instance
    DeepCopy();
  }
  rep->addElement(a, index);
}

// CAN NOT BE INLINED :-(
// Iterator functions
// For this implementation, a copy is made in order to iterate over a GBList
// with >1 instance
template <class T>
GBListIterator<T> GBList<T>::begin() {
  if (rep == NULL) {
    RECORDHERE(rep = new GBListRep<T>();)
    if(rep==NULL)DBG();
  } else if (rep->count() != 1)  {// More than one instance
    DeepCopy();
  }
  return rep->begin();
}

template <class T>
GBListIterator<T> GBList<T>::end() {
  if (rep == NULL) {
    RECORDHERE(rep = new GBListRep<T>();)
    if(rep==NULL)DBG();
  } else if (rep->count() != 1)  {// More than one instance
    DeepCopy();
  }
  return rep->end();
}

// CAN NOT BE INLINED :-(
template<class T>
void GBListRep<T>::push_back(const T & a) {
  ptrTOaList->append(a);
  incrementSize();
};

// CAN NOT BE INLINED :-(
template<class T>
void GBListRep<T>::addElement(const T & a,int index) {
  if(index==1) {
     ptrTOaList->prepend(a);
     incrementSize();
  } else if(index==_size+1) {
     push_back(a);
  } else {
    Pix soFar = ptrTOaList->first();
    const int num = index-2;
    for (int i=1;i<=num;++i) {
      ptrTOaList->next(soFar);
    }
    if(index==_size) {
      ptrTOaList->ins_after(soFar,a);
    }
    else { ptrTOaList->next(soFar);
      ptrTOaList->ins_before(soFar,a);
    }
    incrementSize();
  }
}

// CAN NOT BE INLINED :-(
template<class T>
void GBListRep<T>::removeElement(int index) {
  if(index==1) {
    ptrTOaList->del_front();
  } else {
    Pix pixer = ptrTOaList->first();
    const int num = index-2;
    for (int i=1;i<=num;++i) {
        ptrTOaList->next(pixer);
    }
    ptrTOaList->del_after(pixer);
  }
  decrementSize();
}

// CAN NOT BE INLINED :-(
template<class T>
void GBList<T>::DeepCopy() {
#ifdef DEBUGDEEPCOPY
  if (rep->count() == 1) {
    DBG();
  } else
#endif
  if (rep->count() != 1) // More than one instance
  {
#if 0
GBStream << "Going to do a deep copy.\n";
GBStream << "count is:" << rep->count() << "\n";
#endif
    rep->decrement();
    RECORDHERE(GBListRep<T> * tmp = new GBListRep<T>(*rep);)
    if (tmp == NULL) DBG();
    rep = tmp;
  }
}

// CAN NOT BE INLINED :-(
// This really shouldn't be used anymore; the GBList class will set rep = 0
//   in the case of empty list
template <class T>
void GBListRep<T>::clear() { 
  if(_size>0) {
    ptrTOaList->clear(); 
    zeroSize();
  }
};

template<class T>
GBList<T>::~GBList() {
  if (rep == NULL)
  {  /* Do nothing */ }
  else if (rep->decrement() == 0) {
    RECORDHERE(delete rep;)  // if "quick" empty list, this will fall through
    rep = NULL;
  }
#if 0
  else {
    GBStream << "Deleting a delayed list which was not copied.\n";
  };
#endif
}

// Clear is a special case; it doesn't use DeepCopy but checks for >1 instance
template <class T>
void GBList<T>::clear() {
  if (rep != NULL)
    if (rep->count() != 1) {
      rep->decrement();
      RECORDHERE(rep = new GBListRep<T>();)
      if (rep == NULL) DBG();
    }
    else // Only one instance
    {
      RECORDHERE(delete rep;)
      rep = NULL;
    }
}

template<class T>
GBListRep<T>::GBListRep(const GBListRep<T> & aList) : refcnt(1) {
  RECORDHERE(ptrTOaList = new DLList<T>();)
  if (ptrTOaList == NULL) DBG();
  zeroSize();
  joinTo(aList);
};

template<class T>
GBListRep<T>::~GBListRep() {
  clear();
  RECORDHERE(delete ptrTOaList;)
};

template<class T>
void GBListRep<T>::operator = (const GBListRep<T> & aList) {
  if(this != &aList) {
    refcnt = 1;
    clear();
    d_helper_1 = aList.begin();
    for(int j=1;j<=aList._size;++d_helper_1,++j) {
      push_back(*d_helper_1);
    }
  }
};

template<class T>
void GBListRep<T>::removeDuplicates() {
// change this function very carefully, if at all!!!  
   d_helper_1 = ((const GBListRep<T> *)this)->begin();
   for(int i=1; i<=_size-1;++i,++d_helper_1) {
      d_helper_2 = d_helper_1;
      ++d_helper_2;
      for (int j=i+1; j<=_size;++j,++d_helper_2) {
        if ((*d_helper_1)==(*d_helper_2)) {
          removeElement(j); 
          j--;
          d_helper_2 = ((const GBListRep<T> *)this)->begin(); d_helper_2.advance(j-1);
        }
      }
   }
};


template <class T>
bool GBListRep<T>::operator ==(const GBListRep<T> & RightSide) const {
  bool result = _size == RightSide._size;
  if(result) {
    d_helper_1 = begin();
    d_helper_2 = RightSide.begin();
    for(int k=1;k<=_size && result;++k,++d_helper_1,++d_helper_2) {
      result = (*d_helper_1) == (*d_helper_2); 
    }
  }
  return result;
};

template <class T>
int GBListRep<T>::Ifind(const T & a) const {
  int result = -1;
  d_helper_1 = begin();
  for(int j=1;j<=_size;++j,++d_helper_1) {
    if(a==(*d_helper_1)) {
      result = j;
      break;
    }
  }
  return result;
};

template <class T>
void GBListRep<T>::joinTo(const GBListRep<T> & bList) {
   const int len = bList._size;
   d_helper_1 = bList.begin();
   for(int j=1;j<=len;++d_helper_1,++j) {
     push_back(*d_helper_1);
   }
};

template<class T>
void GBListRep<T>::pop_front() {
  ptrTOaList->del_front();
  decrementSize();
}

template<class U>
const_GBListIterator<U>::const_GBListIterator(
    const GBListRep<U> & ptr,int m) :  _p(0), _ptr(ptr.ptrTOaList) {
  _p = _ptr->first();
  for(int i=1;i<m;++i) {
    _ptr->next(_p);
  }
};

template<class T>
MyOstream & operator <<(MyOstream & os, const const_GBListIterator<T> & iter) {
  os << iter._ptr  << ' '  << iter._p << '\n';
  return os;
};

template<class U>
GBListIterator<U>::~GBListIterator(){};
#endif
