(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      6673,        247]*)
(*NotebookOutlinePosition[      7621,        277]*)
(*  CellTagsIndexPosition[      7577,        273]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["NCGuts Demo", "Title",
  FontFamily->"Application"],

Cell[CellGroupData[{

Cell["Loading the needed files...", "Section",
  FontFamily->"Application"],

Cell["\<\
The following lines may need to be altered depending on where \
NCAlgebra and NCGuts are located on your system:\
\>", "Text",
  FontFamily->"Application"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(<< NCAlgebra.m\), "\[IndentingNewLine]", 
    \(<< NCGuts.m\)}], "Input",
  FontFamily->"Application"],

Cell[BoxData[
    \("You have already loaded NCAlgebra.m"\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Examples depicting the option ",
  StyleBox["NCStrongProduct1",
    FontColor->GrayLevel[0.500008]]
}], "Section",
  FontFamily->"Application"],

Cell["\<\
When NCGuts[NCStrongProduct1\[Rule]False], the following attempt to \
multiply matrices with noncommutative entries using \"**\"fails:\
\>", "Text",\

  FontFamily->"Application"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(NCGuts[NCStrongProduct1 \[Rule] False];\)\), "\[IndentingNewLine]", 
    \({{a, b}, {c, d}} ** {{x}, {y}} == 
      MatMult[{{a, b}, {c, d}}, {{x}, {y}}]\)}], "Input",
  FontFamily->"Application"],

Cell[BoxData[
    \({{a, b}, {c, d}} ** {{x}, {y}} == {{a ** x + b ** y}, {c ** x + 
            d ** y}}\)], "Output"]
}, Open  ]],

Cell["\<\
When NCGuts[NCStrongProduct1\[Rule]True],the following attempt to \
multiply matrices with noncommutative entries using \"**\" is \
successful:\
\>", "Text",
  FontFamily->"Application"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(NCGuts[NCStrongProduct1 \[Rule] True];\)\), "\n", 
    \({{a, b}, {c, d}} ** {{x}, {y}} == 
      MatMult[{{a, b}, {c, d}}, {{x}, {y}}]\)}], "Input",
  FontFamily->"Application"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell["In addition,\"tp[ ]\" and \"tpMat[ ]\" are the same:", "Text",
  FontFamily->"Application"],

Cell[CellGroupData[{

Cell[BoxData[
    \(tp[{{a, b}, {c, d}}] == tpMat[{{a, b}, {c, d}}]\)], "Input",
  FontFamily->"Application"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell["\<\
The default is set to True.  After NCGuts is loaded, \"**\" will \
multiply matrices with noncommutative entries.\
\>", "Text",
  FontFamily->"Application"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Examples depicting the option ",
  StyleBox["NCStrongProduct2",
    FontColor->GrayLevel[0.500008]]
}], "Section",
  FontFamily->"Application"],

Cell[TextData[{
  StyleBox["NCGuts[NCStrongProduct2\[Rule]True] inverts matrices using the \
\"inv[ ]\" command.  ",
    FontWeight->"Plain"],
  Cell[BoxData[
      \(\((The\ limitations\ of\ NCInverse\ remain\ with\ inv[\ ])\)\)], 
    "Input",
    FontFamily->"Application",
    FontWeight->"Plain"],
  StyleBox[":",
    FontWeight->"Plain"]
}], "Text",
  FontFamily->"Application",
  FontWeight->"Bold",
  FontVariations->{"CompatibilityType"->0}],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(NCGuts[NCStrongProduct2 \[Rule] True];\)\), "\[IndentingNewLine]", 
    \(inv[{{a, b}, {0, c}}] // MatrixForm\)}], "Input",
  FontFamily->"Application"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(inv[a]\), \(-inv[a] ** b ** inv[c]\)},
            {"0", \(inv[c]\)}
            }], "\[NoBreak]", ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell["\<\
NCGuts[NCStrongProduct2->True] forces \
NCGuts[NCStrongProduct1->True].  In other words, \
NCGuts[NCStrongProduct2->True] makes \"**\" multiply matrices with \
noncommutative entries.\
\>", "Text",
  FontFamily->"Application"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(NCGuts[NCStrongProduct1 \[Rule] False];\)\), "\[IndentingNewLine]", 
    \(\(NCGuts[NCStrongProduct2 \[Rule] True];\)\), "\[IndentingNewLine]", 
    \({{a, b}, {c, d}} ** {{x}, {y}}\)}], "Input",
  FontFamily->"Application"],

Cell[BoxData[
    \({{a ** x + b ** y}, {c ** x + d ** y}}\)], "Output"]
}, Open  ]],

Cell["The default for NCStrongProduct2 is False.", "Text",
  FontFamily->"Application"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Examples depicting the option ",
  StyleBox["NCSymbols",
    FontColor->GrayLevel[0.500008]]
}], "Section",
  FontFamily->"Application"],

Cell["\<\
NCGuts[NCSymbols\[Rule]True] makes all symbols --\:f220capital \
letters or not -- to be noncommutative.  The default option for this is \
False.\
\>", "Text",
  FontFamily->"Application"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(NCGuts[NCSymbols \[Rule] False];\)\), "\[IndentingNewLine]", 
    \(A ** b ** A\), "\[IndentingNewLine]", 
    \(\(NCGuts[NCSymbols \[Rule] True];\)\), "\[IndentingNewLine]", 
    \(A ** b ** A\)}], "Input",
  FontFamily->"Application"],

Cell[BoxData[
    \(A\^2\ b\)], "Output"],

Cell[BoxData[
    \(A ** b ** A\)], "Output"]
}, Open  ]],

Cell["The default for NCSymbols is False.", "Text",
  FontFamily->"Application"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{880, 776},
WindowMargins->{{Automatic, 55}, {Automatic, 47}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "toal", \
"MAYBE-NC00"}, "NCGutsDemo.nb.ps", CharacterEncoding -> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1739, 51, 57, 1, 98, "Title"],

Cell[CellGroupData[{
Cell[1821, 56, 75, 1, 49, "Section"],
Cell[1899, 59, 165, 4, 30, "Text"],

Cell[CellGroupData[{
Cell[2089, 67, 123, 3, 46, "Input"],
Cell[2215, 72, 70, 1, 23, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[2334, 79, 162, 5, 49, "Section"],
Cell[2499, 86, 189, 5, 30, "Text"],

Cell[CellGroupData[{
Cell[2713, 95, 219, 4, 46, "Input"],
Cell[2935, 101, 119, 2, 27, "Output"]
}, Open  ]],
Cell[3069, 106, 196, 5, 30, "Text"],

Cell[CellGroupData[{
Cell[3290, 115, 201, 4, 46, "Input"],
Cell[3494, 121, 38, 1, 27, "Output"]
}, Open  ]],
Cell[3547, 125, 97, 1, 30, "Text"],

Cell[CellGroupData[{
Cell[3669, 130, 109, 2, 28, "Input"],
Cell[3781, 134, 38, 1, 27, "Output"]
}, Open  ]],
Cell[3834, 138, 166, 4, 30, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4037, 147, 162, 5, 49, "Section"],
Cell[4202, 154, 450, 14, 30, "Text"],

Cell[CellGroupData[{
Cell[4677, 172, 175, 3, 46, "Input"],
Cell[4855, 177, 226, 6, 45, "Output"]
}, Open  ]],
Cell[5096, 186, 236, 6, 46, "Text"],

Cell[CellGroupData[{
Cell[5357, 196, 247, 4, 64, "Input"],
Cell[5607, 202, 72, 1, 27, "Output"]
}, Open  ]],
Cell[5694, 206, 87, 1, 30, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5818, 212, 155, 5, 49, "Section"],
Cell[5976, 219, 198, 5, 30, "Text"],

Cell[CellGroupData[{
Cell[6199, 228, 259, 5, 82, "Input"],
Cell[6461, 235, 41, 1, 28, "Output"],
Cell[6505, 238, 45, 1, 27, "Output"]
}, Open  ]],
Cell[6565, 242, 80, 1, 30, "Text"]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

