(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)



(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     52739,       1247]*)
(*NotebookOutlinePosition[     53515,       1274]*)
(*  CellTagsIndexPosition[     53471,       1270]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["\<\

SINGULAR PERTURBATION
     NCAlgebra NOTEBOOK
           follows CDC99
         by  Dell   and Bill
\
\>", "Title"],

Cell[CellGroupData[{

Cell["The system we study and its control (Sec 2)", "Section"],

Cell[TextData[{
  "\n\n",
  Cell[BoxData[{
      RowBox[{
        RowBox[{
          RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                {\(dx\/dt\)},
                {\(\[Epsilon]  dz\/dt\)}
                }], "\[NegativeThinSpace]", ")"}], " ", "=", " ", 
          RowBox[{
            RowBox[{
              RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                    {\(A\_11\), \(A\_12\)},
                    {\(A\_21\), \(A\_22\)}
                    }], "\[NegativeThinSpace]", ")"}], 
              RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                    {"x"},
                    {"z"}
                    }], "\[NegativeThinSpace]", ")"}]}], " ", "+", " ", 
            RowBox[{
              RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                    {\(B\_1\)},
                    {\(B\_2\)}
                    }], "\[NegativeThinSpace]", ")"}], " ", "u"}]}]}], 
        "\n"}], 
      RowBox[{
        RowBox[{"y", " ", "=", " ", 
          RowBox[{
            RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                  {\(M\_1\), \(M\_2\)}
                  }], "\[NegativeThinSpace]", ")"}], 
            RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                  {"x"},
                  {"z"}
                  }], "\[NegativeThinSpace]", ")"}]}]}], "\n", "\t"}], 
      RowBox[{
        \(Minimize\ \ \ J\  = \ \ 
          \[Integral]\_0\%\[Infinity]\ \ \ y\^T\ \ y\  + \ 
            u\^T\ R\ u\ \ \ \ \ dt\), "\n"}], 
    \(where\ \ R\ is\ a\ positive\ definite\ weight\ matrix . \nIt\ is\ well\ 
        known\ that\ the\ solution\ to\ this\ problem\), 
      RowBox[{\(is\ of\ the\ form\), " ", "\n"}], 
      RowBox[{
        RowBox[{\(u\^*\), " ", "=", "  ", 
          RowBox[{
          \(-\ R\^\(-1\)\), "  ", \(B\^T\), "  ", "K", \((\[Epsilon])\), 
            " ", 
            RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                  {"x"},
                  {"z"}
                  }], "\[NegativeThinSpace]", ")"}]}]}], "\n"}], 
      RowBox[{\(where\ \ K\ is\ the\ solution\ to\ the\ ARE\), "\n"}], 
      RowBox[{"        ", 
        \(K\ A\ \  + \ \(A\^T\) K\  - \ K\ B\ \ R\^\(-1\)\ \ \ B\^T\ K\ \  + 
            \ M\^T\ \ M\  = \ 0\), "\n"}], 
      RowBox[{"with", " ", "\n"}], 
      RowBox[{
        RowBox[{
          RowBox[{"A", " ", "=", "    ", 
            RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                  {\(A\_11\), \(A\_12\)},
                  {\(A\_21\/\[Epsilon]\), \(A\_22\/\[Epsilon]\)}
                  }], "\[NegativeThinSpace]", ")"}]}], ",", "  ", 
          RowBox[{"B", " ", "=", "   ", 
            RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                  {\(B\_1\)},
                  {\(B\_2\/\[Epsilon]\)}
                  }], "\[NegativeThinSpace]", ")"}]}], ",", "  ", "\n", 
          "           ", 
          RowBox[{\(and\ \ \ M\), " ", "=", " ", 
            RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                  {\(M\_1\), \(M\_2\)}
                  }], "\[NegativeThinSpace]", ")"}]}]}], "\n"}], 
      RowBox[{
        \(K\  = \ 
          \(K\_0\  + \ \ \ \[Epsilon]\ K\_1\ \  + \ \[Epsilon]\^2\ \ K\_2\  + 
                \  ...  . \)\), "\n", "\t\t"}]}]],
  "\n"
}], "Text"],

Cell[TextData[{
  "Format K compatibly with the partitioning of A using the notation\nof \
[KKO86].\n\t\tK = ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sum]\+\(i = 0\)\%\[Infinity] \[Epsilon]\^i\)]],
  " ",
  Cell[BoxData[
      FormBox[
        RowBox[{"(", GridBox[{
              {\(k\_\(1  i\)\), \(\[Epsilon]\ k\_\(2  i\)\)},
              {\(\[Epsilon]\ k\_\(2  i\)\%T\), \(\[Epsilon]\ k\_\(3  i\)\)}
              }], ")"}], TraditionalForm]]],
  "\n\n"
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Computer Algebra vs the problem (Sec3)", "Section"],

Cell[CellGroupData[{

Cell["\<\

The ZEROth order term of the Riccati equation (Sec3.1)\
\>", "Subsection"],

Cell[CellGroupData[{

Cell["\<\

Computer Algebra Finds the Basic Equation (Sec 3.1.1)\
\>", "Subsubsection"],

Cell["\<\

We begin by  setting the matrices in our problem to 
be noncommutative.\
\>", "Text"],

Cell["\<\

SetNonCommutative[
\tk10,k20,k30,A11, 
\tA12,A21,A22,B1,B2,M1,M2,R
\t];\
\>", "Input"],

Cell[TextData[{
  "Next we create the first term of the power series \n      \n         K  =  \
 ",
  Cell[BoxData[
      \(\(K\_0\  + \[Epsilon]\ K\_1 + \ \[Epsilon]\^2\ \ K\_2\  + \  ...  . 
        \)\)],
    FontFamily->"Bold"],
  "\n\nIt contains an \[Epsilon]."
}], "Text"],

Cell[CellGroupData[{

Cell["\<\

K0 = {{k10, ep**k20},
     {ep**tp[k20], ep**k30 }};
     
K0//MatrixForm     \
\>", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"k10", \(ep\ k20\)},
            {\(ep\ tp[k20]\), \(ep\ k30\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell["\<\

Next we define the block matrices present in our problem.
\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
A  = {{A11, A12},
      {(1/ep)**A21,(1/ep)**A22}};
B  = {{B1},{(1/ep)**B2}};
Mm = {{M1,M2}}; 

A//MatrixForm 
B//MatrixForm
Mm//MatrixForm\
\>", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"A11", "A12"},
            {\(A21\/ep\), \(A22\/ep\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Output"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"B1"},
            {\(B2\/ep\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Output"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"M1", "M2"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[TextData[{
  "The following ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " input creates a function which \nmakes Riccatis and sets G to be the \
optimal control.\n"
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Riccati[K_] := MatMult[K,A] + 
\tMatMult[tpMat[A],K ] -
\tMatMult[K,B,{{Inv[R]}},tpMat[B],K] +
\tMatMult[tpMat[Mm],Mm]

G = - MatMult[{{Inv[R]}},tpMat[B],K0]\
\>", "Input"],

Cell[BoxData[
    \({{\(-Inv[R]**tp[B1]**k10\) - Inv[R]**tp[B2]**tp[k20], 
        \(-ep\)\ Inv[R]**tp[B1]**k20 - Inv[R]**tp[B2]**k30}}\)], "Output"]
}, Open  ]],

Cell["\<\

\
\>", "Input"],

Cell[TextData[
"Extract the zero-th order terms in \[Epsilon] \nfrom the upper left entry of \
the perturbed\nRiccati  to create a polynomial. "], "Text"],

Cell[CellGroupData[{

Cell["\<\

Ep10 = NCTermsOfDegree[ 
\t\t\tRiccati[K0][[1,1]],{ep},{0} ]\
\>", "Input"],

Cell[BoxData[
    \(k10**A11 + k20**A21 + tp[A11]**k10 + tp[A21]**tp[k20] + tp[M1]**M1 - 
      \((k10**B1 + k20**B2)\)**Inv[R]**tp[B1]**k10 - 
      \((k10**B1 + k20**B2)\)**Inv[R]**tp[B2]**tp[k20]\)], "Output"]
}, Open  ]],

Cell[TextData[
"Extract the zero-th order terms in \[Epsilon] from the upper right entry of \
the perturbed Riccati  to create a polynomial. "], "Text"],

Cell[CellGroupData[{

Cell["\<\
Ep20 = NCTermsOfDegree[ 
\t\t\tRiccati[K0][[1,2]],{ep},{0} ]\
\>", "Input"],

Cell[BoxData[
    \(k10**A12 + k20**A22 + tp[A21]**k30 + tp[M1]**M2 - 
      \((k10**B1 + k20**B2)\)**Inv[R]**tp[B2]**k30\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
NCCollect[ NCExpand[Ep20], {k20}]

\
\>", "Input"],

Cell[BoxData[
    \(k10**A12 + k20**\((A22 - B2**Inv[R]**tp[B2]**k30)\) + tp[A21]**k30 + 
      tp[M1]**M2 - k10**B1**Inv[R]**tp[B2]**k30\)], "Output"]
}, Open  ]],

Cell["\<\
Notice that by assuming the invertibility of  
                               (A22-B2**Inv[R]**tp[B2]**k30)
 in the above equation we can solve for k20.

         k20 = ( k10**A12  +tp[A21]**k30+
               tp[M1]**M2-k10**B1**Inv[R]**tp[B2]**k30 ) **   
               Inv[ A22-B2**Inv[R]**tp[B2]**k30 ]

This observation will be used below.
\
\>", "Text"],

Cell[TextData[
"Next, extract the zero-th order terms in \[Epsilon]  from the \nlower right \
entry of the perturbed Riccati  to create a polynomial. \n"], "Text"],

Cell[CellGroupData[{

Cell["\<\
Ep30 = NCTermsOfDegree[ 
\t\t\tRiccati[K0][[2,2]],{ep},{0} ]
\
\>", "Input"],

Cell[BoxData[
    \(k30**A22 + tp[A22]**k30 + tp[M2]**M2 - k30**B2**Inv[R]**tp[B2]**k30\)], 
  "Output"]
}, Open  ]],

Cell["\<\
Notice that the above equation is a Riccati in k30 and can be used \
to
determine this matrix.

The matrix which is still unknown is k10.  Solving for it by hand is a bit \
tricky, see Section 3.4 [KKO86].
Solving for k10 by computer is what  we turn to after introducing  other \
basic though rinky dink algebraic identities.
\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["    Other Algebraic Identities Known to Hold (Sec 3.1.3)    ", 
  "Subsubsection"],

Cell["\<\

Abbreviations which are standard conventions in the field [KKO86]:\
\>", 
  "Text"],

Cell["\<\

SetNonCommutative[G1,G2,N0,M0,A0,B0,R0];
Abbreviations = {
\tG1 == G[[1,1]]/.ep->0,
\tG2 == G[[1,2]]/.ep->0,
\tN0 == - M2**Inv[A22]**B2,
\tM0 == M1 - M2**Inv[A22]**A21,
\tA0 == A11 - A12 ** Inv[A22]**A21,
\tB0 == B1 - A12**Inv[A22]**B2,
\tR0 == R + tp[N0]**N0
\t};\
\>", "Input"],

Cell["\<\

Also some of our matrices are known to be self adjoint.\
\>", "Text"],

Cell["\<\

SelfAdjoint = {
\tk10 == tp[k10],
\tk30 == tp[k30],
\tR == tp[R],
\tR0 == tp[R0],
\tInv[R] == tp[Inv[R]],
\tInv[R0] == tp[Inv[R0]]
\t};\
\>", "Input"],

Cell["\<\
It is standard that A0, A11, and A22 are invertible.
Also, the fast  sub-system is (assumed ) strictly stable which makes
                   A22 + B2**G2  invertible.  
Recall  the 
              invertibility of    A22 - B2**Inv[R]**tp[B2]**k30 
 follows from our solution for k20.
We make the resulting relations with the next command.\
\>", "Text"],

Cell["\<\

inverses = NCMakeRelations[{Inv,
\tR,R0,A0,A11,A22,(A22 + B2**G2), 
\t(A22 - B2**Inv[R]**tp[B2]**k30) }];\
\>", "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\

Solving for k10 using a  Grobner Basis.\
\>", "Subsubsection"],

Cell["\<\
Next we combine all the relations created above and add in 
their tranposes. 
\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
Relations = Union[{Ep10,Ep20,Ep30},
 \tAbbreviations,
\tSelfAdjoint, inverses ];

AllRelations = NCAddTp[ Relations ]\
\>", "Input"],

Cell[BoxData[
    \({A0 == A11 - A12**Inv[A22]**A21, B0 == B1 - A12**Inv[A22]**B2, 
      G1 == \(-Inv[R]**tp[B1]**k10\) - Inv[R]**tp[B2]**tp[k20], 
      G2 == \(-Inv[R]**tp[B2]**k30\), k10 == tp[k10], k30 == tp[k30], 
      M0 == M1 - M2**Inv[A22]**A21, N0 == \(-M2**Inv[A22]**B2\), R == tp[R], 
      R0 == R + tp[N0]**N0, R0 == tp[R0], Inv[R] == tp[Inv[R]], 
      Inv[R0] == tp[Inv[R0]], A0**Inv[A0] == 1, A11**Inv[A11] == 1, 
      A22**Inv[A22] == 1, R**Inv[R] == 1, R0**Inv[R0] == 1, Inv[A0]**A0 == 1, 
      Inv[A11]**A11 == 1, Inv[A22]**A22 == 1, Inv[R]**R == 1, 
      Inv[R0]**R0 == 1, Inv[A22 + B2**G2]**\((A22 + B2**G2)\) == 1, 
      Inv[A22 - B2**Inv[R]**tp[B2]**k30]**
          \((A22 - B2**Inv[R]**tp[B2]**k30)\) == 1, 
      \((A22 + B2**G2)\)**Inv[A22 + B2**G2] == 1, 
      \((A22 - B2**Inv[R]**tp[B2]**k30)\)**
          Inv[A22 - B2**Inv[R]**tp[B2]**k30] == 1, tp[A0]**tp[Inv[A0]] == 1, 
      tp[A11]**tp[Inv[A11]] == 1, tp[A22]**tp[Inv[A22]] == 1, 
      \((tp[G2]**tp[B2] + tp[A22])\)**tp[Inv[A22 + B2**G2]] == 1, 
      \((\(-tp[k30]**B2**tp[Inv[R]]**tp[B2]\) + tp[A22])\)**
          tp[Inv[A22 - B2**Inv[R]**tp[B2]**k30]] == 1, 
      tp[R]**tp[Inv[R]] == 1, tp[R0]**tp[Inv[R0]] == 1, 
      tp[Inv[A0]]**tp[A0] == 1, tp[Inv[A11]]**tp[A11] == 1, 
      tp[Inv[A22]]**tp[A22] == 1, tp[Inv[R]]**tp[R] == 1, 
      tp[Inv[R0]]**tp[R0] == 1, 
      tp[Inv[A22 + B2**G2]]**\((tp[G2]**tp[B2] + tp[A22])\) == 1, 
      tp[Inv[A22 - B2**Inv[R]**tp[B2]**k30]]**
          \((\(-tp[k30]**B2**tp[Inv[R]]**tp[B2]\) + tp[A22])\) == 1, 
      tp[A0] == \(-tp[A21]**tp[Inv[A22]]**tp[A12]\) + tp[A11], 
      tp[B0] == \(-tp[B2]**tp[Inv[A22]]**tp[A12]\) + tp[B1], 
      tp[G1] == \(-k20**B2**tp[Inv[R]]\) - tp[k10]**B1**tp[Inv[R]], 
      tp[G2] == \(-tp[k30]**B2**tp[Inv[R]]\), tp[k10] == k10, tp[k30] == k30, 
      tp[M0] == \(-tp[A21]**tp[Inv[A22]]**tp[M2]\) + tp[M1], 
      tp[N0] == \(-tp[B2]**tp[Inv[A22]]**tp[M2]\), tp[R] == R, tp[R0] == R0, 
      tp[R0] == tp[N0]**N0 + tp[R], tp[Inv[R]] == Inv[R], 
      tp[Inv[R0]] == Inv[R0], 
      k10**A12 + k20**A22 + tp[A21]**k30 + tp[M1]**M2 - 
        \((k10**B1 + k20**B2)\)**Inv[R]**tp[B2]**k30, 
      k10**A11 + k20**A21 + tp[A11]**k10 + tp[A21]**tp[k20] + tp[M1]**M1 - 
        \((k10**B1 + k20**B2)\)**Inv[R]**tp[B1]**k10 - 
        \((k10**B1 + k20**B2)\)**Inv[R]**tp[B2]**tp[k20], 
      k20**A21 + tp[A11]**tp[k10] + tp[A21]**tp[k20] + tp[k10]**A11 + 
        tp[M1]**M1 - 
        k20**B2**tp[Inv[R]]**\((tp[B1]**tp[k10] + tp[B2]**tp[k20])\) - 
        tp[k10]**B1**tp[Inv[R]]**\((tp[B1]**tp[k10] + tp[B2]**tp[k20])\), 
      tp[A12]**tp[k10] + tp[A22]**tp[k20] + tp[k30]**A21 + tp[M2]**M1 - 
        tp[k30]**B2**tp[Inv[R]]**\((tp[B1]**tp[k10] + tp[B2]**tp[k20])\), 
      k30**A22 + tp[A22]**k30 + tp[M2]**M2 - k30**B2**Inv[R]**tp[B2]**k30, 
      tp[A22]**tp[k30] + tp[k30]**A22 + tp[M2]**M2 - 
        tp[k30]**B2**tp[Inv[R]]**tp[B2]**tp[k30]}\)], "Output"]
}, Open  ]],

Cell["\<\
Now that we have created all equations which need to be solved 
simultaneously to find k10,  we invoke the NC Groebner Basis
program NCProccess to solve for  k10. First we set the order reflecting our \
priorities on which variables should be eliminated.
\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\

NCAutomaticOrder[ {{N0,M0,R0,A0,B0}, 
\t\t{k10},{B1,B2,M1,M2,R,
\t\tA11,A12,A21,A22,
\t\tInv[A22-B2**Inv[R]**tp[B2]**k30],
\t    tp[Inv[A22-B2**Inv[R]**tp[B2]**k30]]},
\t    {k30},{k20},{G1,G2}},  AllRelations];
\
\>", "Input"],

Cell[BoxData[
    InterpretationBox[
      \(" Computed monomial order is : "\[InvisibleSpace]{{N0, tp[N0], M0, 
            tp[M0], R0, tp[Inv[R0]], tp[R0], Inv[R0], A0, tp[Inv[A0]], 
            tp[A0], Inv[A0], B0, tp[B0]}, {k10, tp[k10]}, {B1, tp[B1], B2, 
            tp[B2], M1, tp[M1], M2, tp[M2], R, tp[Inv[R]], tp[R], Inv[R], 
            A11, tp[Inv[A11]], tp[A11], Inv[A11], A12, tp[A12], A21, tp[A21], 
            A22, tp[Inv[A22 + B2**G2]], tp[Inv[A22]], tp[A22], 
            Inv[A22 + B2**G2], Inv[A22], Inv[A22 - B2**Inv[R]**tp[B2]**k30], 
            tp[Inv[A22 - B2**Inv[R]**tp[B2]**k30]]}, {k30, tp[k30]}, {k20, 
            tp[k20]}, {G1, tp[G1], G2, tp[G2]}}\),
      SequenceForm[ " Computed monomial order is : ", {{N0, 
        tp[ N0], M0, 
        tp[ M0], R0, 
        tp[ 
          Inv[ R0]], 
        tp[ R0], 
        Inv[ R0], A0, 
        tp[ 
          Inv[ A0]], 
        tp[ A0], 
        Inv[ A0], B0, 
        tp[ B0]}, {k10, 
        tp[ k10]}, {B1, 
        tp[ B1], B2, 
        tp[ B2], M1, 
        tp[ M1], M2, 
        tp[ M2], R, 
        tp[ 
          Inv[ R]], 
        tp[ R], 
        Inv[ R], A11, 
        tp[ 
          Inv[ A11]], 
        tp[ A11], 
        Inv[ A11], A12, 
        tp[ A12], A21, 
        tp[ A21], A22, 
        tp[ 
          Inv[ 
            Plus[ A22, 
              NonCommutativeMultiply[ B2, G2]]]], 
        tp[ 
          Inv[ A22]], 
        tp[ A22], 
        Inv[ 
          Plus[ A22, 
            NonCommutativeMultiply[ B2, G2]]], 
        Inv[ A22], 
        Inv[ 
          Plus[ A22, 
            Times[ -1, 
              NonCommutativeMultiply[ B2, 
                Inv[ R], 
                tp[ B2], k30]]]], 
        tp[ 
          Inv[ 
            Plus[ A22, 
              Times[ -1, 
                NonCommutativeMultiply[ B2, 
                  Inv[ R], 
                  tp[ B2], k30]]]]]}, {k30, 
        tp[ k30]}, {k20, 
        tp[ k20]}, {G1, 
        tp[ G1], G2, 
        tp[ G2]}}],
      Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
  "\nFinally we make the call to NCProcess[ ]  which will \ncreate a sorted \
Groebner basis from the relations in AllRelations\nand hopefully give us an \
answer for k10.  The ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " answer\ncan be found below and the TeX answer can be found on the \nmain \
Singular Perturbation page.\n"
}], "Text"],

Cell[CellGroupData[{

Cell["\<\

answer = NCProcess[AllRelations,4,
\t\"FindK10\",SBByCat->False  ,
\tDegreeCap->14, DegreeSumCap->17] ;\
\>", "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
      "\<\"Done outputting results to the stream \"\>", "\[InvisibleSpace]", 
        InterpretationBox[
          StyleBox[\(OutputStream["\<FindK10.tex\>", \ 6]\),
            ShowStringCharacters->True,
            NumberMarks->True],
          Format[ 
            OutputStream[ "FindK10.tex", 6], InputForm],
          Editable->True,
          AutoDelete->True]}],
      SequenceForm[ "Done outputting results to the stream ", 
        Format[ 
          OutputStream[ "FindK10.tex", 6], InputForm]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\

answer\
\>", "Input"],

Cell[BoxData[
    \({{}, {A0**Inv[A0] \[Rule] 1, R0**tp[Inv[R0]] \[Rule] 1, 
        Inv[A0]**A0 \[Rule] 1, tp[A0]**tp[Inv[A0]] \[Rule] 1, 
        tp[Inv[A0]]**tp[A0] \[Rule] 1, tp[Inv[R0]]**R0 \[Rule] 1, 
        G1 \[Rule] 
          \(-tp[Inv[R0]]**tp[B0]**k10\) + 
            tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**k30**A21 + 
            tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**tp[M2]**M1 - 
            tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**
              tp[B0]**k10 + 
            tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**tp[M2]**N0**
              tp[Inv[R0]]**tp[B0]**k10, 
        G2 \[Rule] \(-tp[Inv[R]]**tp[B2]**k30\), 
        k20 \[Rule] 
          \(-k10**A12**Inv[A22]\) + tp[A21]**tp[Inv[A22]]**k30 - 
            tp[M0]**M2**Inv[A22 + B2**G2] - 
            k10**B0**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30 + 
            k10**B0**tp[Inv[R0]]**tp[N0]**M2**Inv[A22], 
        R \[Rule] R0 - tp[N0]**N0, Inv[R] \[Rule] tp[Inv[R]], 
        Inv[R0] \[Rule] tp[Inv[R0]], 
        Inv[A22 - B2**Inv[R]**tp[B2]**k30] \[Rule] Inv[A22 + B2**G2], 
        tp[G1] \[Rule] 
          \(-k10**B0**tp[Inv[R0]]\) - 
            tp[A21]**tp[Inv[A22]]**k30**B2**tp[Inv[R]] + 
            tp[M0]**M2**Inv[A22 + B2**G2]**B2**tp[Inv[R]] + 
            k10**B0**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30**B2**tp[Inv[R]], 
        tp[G2] \[Rule] \(-k30**B2**tp[Inv[R]]\), tp[k10] \[Rule] k10, 
        tp[k20] \[Rule] 
          \(-tp[Inv[A22]]**tp[A12]**k10\) - tp[Inv[A22 + B2**G2]]**k30**A21 - 
            tp[Inv[A22 + B2**G2]]**tp[M2]**M1 + 
            tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**tp[B0]**k10 + 
            tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**tp[B0]**k10 - 
            tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]]**tp[B0]**k10, 
        tp[k30] \[Rule] k30, tp[R] \[Rule] R0 - tp[N0]**N0, 
        tp[R0] \[Rule] R0, 
        tp[Inv[A22 - B2**Inv[R]**tp[B2]**k30]] \[Rule] 
          tp[Inv[A22 + B2**G2]]}, {
        \(-Inv[A22 + B2**G2]\) + Inv[A22 - B2**Inv[R]**tp[B2]**k30], 
        \(-1\) + A0**Inv[A0], \(-1\) + A11**Inv[A11], \(-1\) + A22**Inv[A22], 
        \(-1\) + R0**tp[Inv[R0]], \(-1\) + Inv[A0]**A0, 
        \(-1\) + Inv[A11]**A11, \(-1\) + Inv[A22]**A22, 
        \(-1\) + tp[A0]**tp[Inv[A0]], \(-1\) + tp[A11]**tp[Inv[A11]], 
        \(-1\) + tp[A22]**tp[Inv[A22]], R - R0 + tp[N0]**N0, 
        \(-1\) + tp[Inv[A0]]**tp[A0], \(-1\) + tp[Inv[A11]]**tp[A11], 
        \(-1\) + tp[Inv[A22]]**tp[A22], \(-1\) + tp[Inv[R0]]**R0, 
        A0 - A11 + A12**Inv[A22]**A21, B0 - B1 + A12**Inv[A22]**B2, 
        M0 - M1 + M2**Inv[A22]**A21, N0 + M2**Inv[A22]**B2, 
        1 - R0**tp[Inv[R]] + tp[N0]**N0**tp[Inv[R]], 
        G2 + tp[Inv[R]]**tp[B2]**k30, 
        1 - tp[Inv[R]]**R0 + tp[Inv[R]]**tp[N0]**N0, 
        k30 + tp[Inv[A22 + B2**G2]]**k30**A22 + 
          tp[Inv[A22 + B2**G2]]**tp[M2]**M2, 
        k30**Inv[A22 + B2**G2] + tp[Inv[A22]]**k30 + 
          tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2], 
        tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**k30 - 
          tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2] - 
          tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30 + 
          tp[Inv[R0]]**tp[N0]**M2**Inv[A22], 
        \(-tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]\) + 
          tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R]] - 
          tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]] + 
          tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]], 
        \(-A12**Inv[A22]\) + A12**Inv[A22 + B2**G2] - 
          B0**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30 + 
          B0**tp[Inv[R0]]**tp[N0]**M2**Inv[A22] + 
          B1**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30 - 
          B1**tp[Inv[R0]]**tp[N0]**M2**Inv[A22], 
        \(-1\) + A22**Inv[A22 + B2**G2] + 
          B2**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30 - 
          B2**tp[Inv[R0]]**tp[N0]**M2**Inv[A22], 
        \(-k30**A22\) - tp[A22]**k30 - tp[M2]**M2 + 
          k30**B2**tp[Inv[R]]**tp[B2]**k30, 
        1 - tp[A22]**tp[Inv[A22 + B2**G2]] + 
          k30**B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]], 
        M2**Inv[A22] - M2**Inv[A22 + B2**G2] + 
          N0**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30 - 
          N0**tp[Inv[R0]]**tp[N0]**M2**Inv[A22], 
        1 - Inv[A22 + B2**G2]**A22 + 
          Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**k30, 
        \(-k10**A0\) - tp[A0]**k10 - tp[M0]**M0 + 
          k10**B0**tp[Inv[R0]]**tp[B0]**k10 + 
          k10**B0**tp[Inv[R0]]**tp[N0]**M0 + 
          tp[M0]**N0**tp[Inv[R0]]**tp[B0]**k10 + 
          tp[M0]**N0**tp[Inv[R0]]**tp[N0]**M0, 
        tp[Inv[A22]]**tp[A12] - tp[Inv[A22 + B2**G2]]**tp[A12] - 
          tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**tp[B0] + 
          tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**tp[B1] - 
          tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**tp[B0] + 
          tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**tp[B1] + 
          tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]]**tp[B0] - 
          tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]]**tp[B1], 
        1 - tp[Inv[A22 + B2**G2]]**tp[A22] + 
          tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**tp[B2] + 
          tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**tp[B2] - 
          tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]]**tp[B2], 
        \(-tp[Inv[A22]]**tp[M2]\) + tp[Inv[A22 + B2**G2]]**tp[M2] + 
          tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**tp[N0] + 
          tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**tp[N0] - 
          tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]]**tp[N0], 
        k20 + k10**A12**Inv[A22] - tp[A21]**tp[Inv[A22]]**k30 + 
          tp[M0]**M2**Inv[A22 + B2**G2] + 
          k10**B0**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30 - 
          k10**B0**tp[Inv[R0]]**tp[N0]**M2**Inv[A22], 
        \(-A12**Inv[A22]**Inv[A22]**B2**tp[Inv[R]]\) + 
          A12**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]] + 
          A12**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R]] - 
          A12**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**N0**tp[Inv[R]] + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R]] - 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**N0**tp[Inv[R]] - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R]] + 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]], 
        \(-Inv[A22]**B2**tp[Inv[R]]\) + Inv[A22]**B2**tp[Inv[R0]] + 
          A22**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R]] - 
          A22**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**N0**tp[Inv[R]] - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R]] + 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]], 
        M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R]] - 
          M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]] - 
          M2**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R]] + 
          M2**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**N0**tp[Inv[R]] - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R]] + 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]], 
        tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[M2]**N0**
            tp[Inv[R]] + tp[N0]**M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R]] - 
          tp[N0]**M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**N0**tp[Inv[R]] - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R]] + 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]], 
        G1 + tp[Inv[R0]]**tp[B0]**k10 - 
          tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**k30**A21 - 
          tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**tp[M2]**M1 + 
          tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**
            tp[B0]**k10 - 
          tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]]**
            tp[B0]**k10, 
        \(-A12**Inv[A22]**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]\) + 
          A12**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**Inv[A22 + B2**G2] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            k30 + B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            k30 - B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[M2]**M2**Inv[A22 + B2**G2] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[M2]**M2**Inv[A22 + B2**G2], 
        \(-A12**Inv[A22]**Inv[A22 + B2**G2]**Inv[A22]\) + 
          A12**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**Inv[A22] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**k30 + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            Inv[A22] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            Inv[A22] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        \(-A12**Inv[A22]**Inv[A22]**Inv[A22 + B2**G2]\) + 
          A12**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22 + B2**G2] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**k30 + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**
            Inv[A22 + B2**G2] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**k30 - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**
            Inv[A22 + B2**G2] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22 + B2**G2] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22 + B2**G2] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2], 
        \(-A12**Inv[A22]**Inv[A22]**Inv[A22]\) + 
          A12**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**k30 + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        A12**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B0] - 
          A12**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B1] - 
          A12**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B0] + 
          A12**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B1] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[A12] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A12] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[A12] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A12] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B0] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B1] - 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B0] + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B1] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B0] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B1] + 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B0] - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B1], 
        \(-B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]\) + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]] - 
          A12**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2] + 
          A12**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A22] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A22] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B2] + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B2] - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2], 
        \(-A12**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[N0]\) + 
          A12**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[N0] + 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2] - 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2] - 
          B0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[N0] + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[N0] + 
          B1**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[N0] - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[N0], 
        \(-Inv[A22 + B2**G2]**Inv[A22 + B2**G2]\) + 
          A22**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            k30 - B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[M2]**M2**Inv[A22 + B2**G2], 
        \(-Inv[A22 + B2**G2]**Inv[A22]\) + 
          A22**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**Inv[A22] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            Inv[A22] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        \(-Inv[A22]**Inv[A22 + B2**G2]\) + 
          A22**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22 + B2**G2] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**k30 - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**
            Inv[A22 + B2**G2] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22 + B2**G2] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2], 
        \(-Inv[A22]**Inv[A22]\) + 
          A22**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        Inv[A22]**B2**tp[Inv[R0]]**tp[B0] - 
          Inv[A22]**B2**tp[Inv[R0]]**tp[B1] - 
          A22**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B0] + 
          A22**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B1] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[A12] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A12] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B0] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B1] + 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B0] - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B1], 
        B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]] - 
          Inv[A22]**B2**tp[Inv[R0]]**tp[B2] + 
          A22**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A22] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B2] - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2], 
        \(-Inv[A22]**B2**tp[Inv[R0]]**tp[N0]\) + 
          A22**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[N0] - 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2] + 
          B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[N0] - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[N0], 
        M2**Inv[A22]**Inv[A22 + B2**G2]**Inv[A22 + B2**G2] - 
          M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            k30 - N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[M2]**M2**Inv[A22 + B2**G2], 
        M2**Inv[A22]**Inv[A22 + B2**G2]**Inv[A22] - 
          M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**Inv[A22] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            Inv[A22] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        M2**Inv[A22]**Inv[A22]**Inv[A22 + B2**G2] - 
          M2**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22 + B2**G2] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**k30 - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**
            Inv[A22 + B2**G2] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22 + B2**G2] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2], 
        M2**Inv[A22]**Inv[A22]**Inv[A22] - 
          M2**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        \(-M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B0]\) + 
          M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B1] + 
          M2**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B0] - 
          M2**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B1] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[A12] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A12] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B0] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B1] + 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B0] - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B1], 
        N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]] + 
          M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2] - 
          M2**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A22] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B2] - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2], 
        M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[N0] - 
          M2**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[N0] - 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2] + 
          N0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[N0] - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[N0], 
        \(-tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**k30\) + 
          tp[N0]**M2**Inv[A22]**Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            k30 - R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2]**
            Inv[A22 + B2**G2] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22 + B2**G2] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[M2]**M2**Inv[A22 + B2**G2], 
        \(-tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**k30\) + 
          tp[N0]**M2**Inv[A22]**Inv[A22 + B2**G2]**Inv[A22] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            Inv[A22] + 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2]**
            Inv[A22] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[M2]**
            M2**Inv[A22] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**M2**
            Inv[A22 + B2**G2]**Inv[A22] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        \(-tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[Inv[A22]]**k30\) + 
          tp[N0]**M2**Inv[A22]**Inv[A22]**Inv[A22 + B2**G2] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**k30 - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**
            Inv[A22 + B2**G2] + 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22]**
            Inv[A22 + B2**G2] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[Inv[A22]]**tp[M2]**
            M2**Inv[A22 + B2**G2] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22 + B2**G2] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22]]**tp[M2]**M2**Inv[A22 + B2**G2], 
        \(-tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
              tp[Inv[A22 + B2**G2]]**k30\) + 
          tp[N0]**M2**Inv[A22]**Inv[A22]**Inv[A22] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**k30 - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**Inv[A22] + 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22]**
            Inv[A22] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2]**M2**Inv[A22]**Inv[A22] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[Inv[A22 + B2**G2]]**tp[M2]**M2**Inv[A22], 
        \(-tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[A12]\) + 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[A12] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[A12] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A12] + 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**
            tp[B0] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**
            tp[B1] - 
          tp[N0]**M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B0] + 
          tp[N0]**M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B1] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B0] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B1] + 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B0] - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B1], 
        \(-tp[B2]**tp[Inv[A22]]\) + R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]] + 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[A22] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[A22] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**
            tp[B2] + 
          tp[N0]**M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[B2] - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2], 
        tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**tp[M2] - 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22 + B2**G2]]**
            tp[M2] - 
          tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**
            tp[N0] + 
          tp[N0]**M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[N0] + 
          R0**tp[Inv[R]]**tp[B2]**tp[Inv[A22]]**tp[Inv[A22]]**tp[M2]**N0**
            tp[Inv[R0]]**tp[N0] - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[N0], 
        \(-A12**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]\) + 
          A12**Inv[A22]**Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**
            tp[Inv[A22 + B2**G2]] + 
          A12**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2]**
            tp[Inv[A22]] - 
          A12**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**
            tp[Inv[A22 + B2**G2]] + 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2]**tp[Inv[A22]] - 
          B0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]] - 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2]**tp[Inv[A22]] + 
          B1**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]], 
        \(-Inv[A22]**B2**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]\) + 
          Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]] + 
          A22**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2]**
            tp[Inv[A22]] - 
          A22**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**
            tp[Inv[A22 + B2**G2]] - 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2]**tp[Inv[A22]] + 
          B2**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]], 
        M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]] - 
          M2**Inv[A22]**Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**
            tp[Inv[A22 + B2**G2]] - 
          M2**Inv[A22 + B2**G2]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2]**
            tp[Inv[A22]] + 
          M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**
            tp[Inv[A22 + B2**G2]] - 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2]**tp[Inv[A22]] + 
          N0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]], 
        tp[N0]**M2**Inv[A22]**Inv[A22]**B2**tp[Inv[R0]]**tp[B2]**
            tp[Inv[A22]] - 
          tp[N0]**M2**Inv[A22]**Inv[A22 + B2**G2]**B2**tp[Inv[R]]**tp[B2]**
            tp[Inv[A22 + B2**G2]] - 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22]**B2**
            tp[Inv[R0]]**tp[B2]**tp[Inv[A22]] + 
          R0**tp[Inv[R]]**tp[N0]**M2**Inv[A22 + B2**G2]**Inv[A22 + B2**G2]**
            B2**tp[Inv[R]]**tp[B2]**tp[Inv[A22 + B2**G2]], 
        tp[A21]**tp[Inv[A22]]**tp[A12] + tp[A0] - tp[A11], 
        tp[B2]**tp[Inv[A22]]**tp[A12] + tp[B0] - tp[B1], 
        k10**B0**tp[Inv[R0]] + tp[A21]**tp[Inv[A22]]**k30**B2**tp[Inv[R]] - 
          tp[M0]**M2**Inv[A22 + B2**G2]**B2**tp[Inv[R]] - 
          k10**B0**tp[Inv[R0]]**tp[B2]**tp[Inv[A22]]**k30**B2**tp[Inv[R]] + 
          tp[G1], k30**B2**tp[Inv[R]] + tp[G2], \(-k10\) + tp[k10], 
        tp[Inv[A22]]**tp[A12]**k10 + tp[Inv[A22 + B2**G2]]**k30**A21 + 
          tp[Inv[A22 + B2**G2]]**tp[M2]**M1 - 
          tp[Inv[A22]]**tp[M2]**N0**tp[Inv[R0]]**tp[B0]**k10 - 
          tp[Inv[A22 + B2**G2]]**k30**B2**tp[Inv[R0]]**tp[B0]**k10 + 
          tp[Inv[A22 + B2**G2]]**tp[M2]**N0**tp[Inv[R0]]**tp[B0]**k10 + 
          tp[k20], \(-k30\) + tp[k30], 
        tp[A21]**tp[Inv[A22]]**tp[M2] + tp[M0] - tp[M1], 
        tp[B2]**tp[Inv[A22]]**tp[M2] + tp[N0], \(-R0\) + tp[N0]**N0 + tp[R], 
        \(-R0\) + tp[R0], Inv[R] - tp[Inv[R]], Inv[R0] - tp[Inv[R0]], 
        \(-tp[Inv[A22 + B2**G2]]\) + tp[Inv[A22 - B2**Inv[R]**tp[B2]**k30]]}}
      \)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{670, 657},
WindowMargins->{{156, Automatic}, {Automatic, 82}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 126, 7, 324, "Title"],

Cell[CellGroupData[{
Cell[1882, 62, 62, 0, 52, "Section"],
Cell[1947, 64, 3256, 78, 546, "Text"],
Cell[5206, 144, 475, 13, 117, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5718, 162, 57, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[5800, 166, 85, 3, 62, "Subsection"],

Cell[CellGroupData[{
Cell[5910, 173, 87, 3, 58, "Subsubsection"],
Cell[6000, 178, 96, 4, 68, "Text"],
Cell[6099, 184, 97, 6, 87, "Input"],
Cell[6199, 192, 279, 8, 104, "Text"],

Cell[CellGroupData[{
Cell[6503, 204, 104, 6, 87, "Input"],
Cell[6610, 212, 191, 6, 44, "Output"]
}, Open  ]],
Cell[6816, 221, 83, 4, 68, "Text"],

Cell[CellGroupData[{
Cell[6924, 229, 164, 9, 132, "Input"],
Cell[7091, 240, 181, 6, 50, "Output"],
Cell[7275, 248, 159, 6, 50, "Output"],
Cell[7437, 256, 139, 5, 28, "Output"]
}, Open  ]],
Cell[7591, 264, 194, 6, 68, "Text"],

Cell[CellGroupData[{
Cell[7810, 274, 182, 7, 102, "Input"],
Cell[7995, 283, 149, 2, 43, "Output"]
}, Open  ]],
Cell[8159, 288, 26, 3, 42, "Input"],
Cell[8188, 293, 154, 2, 68, "Text"],

Cell[CellGroupData[{
Cell[8367, 299, 86, 4, 57, "Input"],
Cell[8456, 305, 212, 3, 59, "Output"]
}, Open  ]],
Cell[8683, 311, 152, 2, 32, "Text"],

Cell[CellGroupData[{
Cell[8860, 317, 85, 3, 42, "Input"],
Cell[8948, 322, 135, 2, 43, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9120, 329, 60, 4, 57, "Input"],
Cell[9183, 335, 151, 2, 43, "Output"]
}, Open  ]],
Cell[9349, 340, 371, 11, 194, "Text"],
Cell[9723, 353, 163, 2, 68, "Text"],

Cell[CellGroupData[{
Cell[9911, 359, 86, 4, 57, "Input"],
Cell[10000, 365, 104, 2, 27, "Output"]
}, Open  ]],
Cell[10119, 370, 351, 10, 122, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10507, 385, 88, 1, 42, "Subsubsection"],
Cell[10598, 388, 94, 4, 50, "Text"],
Cell[10695, 394, 290, 12, 177, "Input"],
Cell[10988, 408, 80, 3, 50, "Text"],
Cell[11071, 413, 161, 10, 147, "Input"],
Cell[11235, 425, 361, 8, 140, "Text"],
Cell[11599, 435, 131, 5, 72, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[11767, 445, 73, 3, 58, "Subsubsection"],
Cell[11843, 450, 102, 4, 68, "Text"],

Cell[CellGroupData[{
Cell[11970, 458, 142, 6, 87, "Input"],
Cell[12115, 466, 2941, 47, 619, "Output"]
}, Open  ]],
Cell[15071, 516, 279, 6, 104, "Text"],

Cell[CellGroupData[{
Cell[15375, 526, 238, 9, 132, "Input"],
Cell[15616, 537, 2030, 66, 119, "Print"]
}, Open  ]],
Cell[17661, 606, 369, 8, 140, "Text"],

Cell[CellGroupData[{
Cell[18055, 618, 129, 5, 72, "Input"],
Cell[18187, 625, 592, 15, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18816, 645, 32, 3, 42, "Input"],
Cell[18851, 650, 33824, 590, 7051, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

